from __future__ import annotations

import logging
import typing as tp
from functools import cached_property

import grpc
from hist import Hist
import json
import uhi.io.json

from histserv.protos import hist_pb2, hist_pb2_grpc
from histserv.serialize import serialize, deserialize

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Client:
    def __init__(self, address: str) -> None:
        self.address = address

    def __getstate__(self):
        state = dict(self.__dict__)
        state.pop("channel", None)
        return state

    def __enter__(self) -> Client:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        del exc_type, exc_value, traceback  # unused
        self.channel.close()

    @cached_property
    def channel(self) -> grpc.Channel:
        return grpc.insecure_channel(
            self.address,
            # compression=grpc.Compression.Gzip,
            compression=grpc.Compression.NoCompression,  # turn off for now, we compress with numcodecs the byte buffer
            options=[
                ("grpc.max_send_message_length", 1 << 29),
            ],
        )

    @property
    def stub(self) -> hist_pb2_grpc.HistogrammerServiceStub:
        return hist_pb2_grpc.HistogrammerServiceStub(self.channel)

    def init(self, hist: Hist) -> RemoteHist:
        if not isinstance(hist, Hist):
            raise ValueError(f"`hist` must of type `hist.Hist`, got {type(hist)=}")

        # TODO: replace with metadata only serialization once this is resolved:
        # https://github.com/scikit-hep/boost-histogram/issues/1089
        json_obj = json.dumps(hist, default=uhi.io.json.default)
        request = hist_pb2.InitRequest(hist_json=json_obj)

        ret = self.stub.Init(request)
        if not ret.success:
            raise ValueError(ret.message)

        # create a remote hist that one can interact with
        return RemoteHist(client=self, hist_id=ret.message)


class RemoteHist:
    __slots__ = ("_client", "_hist_id")

    def __init__(self, client: Client, hist_id: str) -> None:
        self._client = client
        self._hist_id = hist_id

    @property
    def client(self) -> Client:
        return self._client

    @property
    def hist_id(self) -> str:
        return self._hist_id

    def __repr__(self) -> str:
        return f"RemoteHist<ID={self.hist_id} @{self.client.address}>"

    def fill(self, **kwargs: tp.Any) -> hist_pb2.FillResponse:
        serialized_kwargs = {key: serialize(value) for key, value in kwargs.items()}
        request = hist_pb2.FillRequest(hist_id=self.hist_id, kwargs=serialized_kwargs)
        return self.client.stub.Fill(request)

    def snapshot(self, drop_from_server: bool = False) -> Hist:
        request = hist_pb2.SnapShotRequest(
            hist_id=self.hist_id, drop_from_server=drop_from_server
        )
        ret = self.client.stub.SnapShot(request)
        if not ret.success:
            raise ValueError(ret.message)

        hist_json = json.loads(ret.hist_json)
        content_ser = ret.data
        content = {k: deserialize(v) for k, v in content_ser.items()}

        hist_json["storage"].update(content)
        return Hist(hist_json)

    def flush(self, destination: str = "hist.h5") -> hist_pb2.FlushResponse:
        request = hist_pb2.FlushRequest(hist_id=self.hist_id, destination=destination)
        return self.client.stub.Flush(request)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc
import warnings

import histserv.protos.hist_pb2 as hist__pb2

GRPC_GENERATED_VERSION = "1.78.0"
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower

    _version_not_supported = first_version_is_lower(
        GRPC_VERSION, GRPC_GENERATED_VERSION
    )
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f"The grpc package installed is at version {GRPC_VERSION},"
        + " but the generated code in hist_pb2_grpc.py depends on"
        + f" grpcio>={GRPC_GENERATED_VERSION}."
        + f" Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}"
        + f" or downgrade your generated code using grpcio-tools<={GRPC_VERSION}."
    )


class HistogrammerServiceStub(object):
    """The histogram service definition."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Init = channel.unary_unary(
            "/HistogrammerService/Init",
            request_serializer=hist__pb2.InitRequest.SerializeToString,
            response_deserializer=hist__pb2.InitResponse.FromString,
            _registered_method=True,
        )
        self.Fill = channel.unary_unary(
            "/HistogrammerService/Fill",
            request_serializer=hist__pb2.FillRequest.SerializeToString,
            response_deserializer=hist__pb2.FillResponse.FromString,
            _registered_method=True,
        )
        self.SnapShot = channel.unary_unary(
            "/HistogrammerService/SnapShot",
            request_serializer=hist__pb2.SnapShotRequest.SerializeToString,
            response_deserializer=hist__pb2.SnapShotResponse.FromString,
            _registered_method=True,
        )
        self.Flush = channel.unary_unary(
            "/HistogrammerService/Flush",
            request_serializer=hist__pb2.FlushRequest.SerializeToString,
            response_deserializer=hist__pb2.FlushResponse.FromString,
            _registered_method=True,
        )


class HistogrammerServiceServicer(object):
    """The histogram service definition."""

    def Init(self, request, context):
        """Sends a histogram initialization request"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Fill(self, request, context):
        """Sends a filling request"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SnapShot(self, request, context):
        """Sends a snapshot request of the current histogram"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Flush(self, request, context):
        """Sends a flushing request (to disk)"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_HistogrammerServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Init": grpc.unary_unary_rpc_method_handler(
            servicer.Init,
            request_deserializer=hist__pb2.InitRequest.FromString,
            response_serializer=hist__pb2.InitResponse.SerializeToString,
        ),
        "Fill": grpc.unary_unary_rpc_method_handler(
            servicer.Fill,
            request_deserializer=hist__pb2.FillRequest.FromString,
            response_serializer=hist__pb2.FillResponse.SerializeToString,
        ),
        "SnapShot": grpc.unary_unary_rpc_method_handler(
            servicer.SnapShot,
            request_deserializer=hist__pb2.SnapShotRequest.FromString,
            response_serializer=hist__pb2.SnapShotResponse.SerializeToString,
        ),
        "Flush": grpc.unary_unary_rpc_method_handler(
            servicer.Flush,
            request_deserializer=hist__pb2.FlushRequest.FromString,
            response_serializer=hist__pb2.FlushResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "HistogrammerService", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers("HistogrammerService", rpc_method_handlers)


# This class is part of an EXPERIMENTAL API.
class HistogrammerService(object):
    """The histogram service definition."""

    @staticmethod
    def Init(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/HistogrammerService/Init",
            hist__pb2.InitRequest.SerializeToString,
            hist__pb2.InitResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )

    @staticmethod
    def Fill(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/HistogrammerService/Fill",
            hist__pb2.FillRequest.SerializeToString,
            hist__pb2.FillResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )

    @staticmethod
    def SnapShot(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/HistogrammerService/SnapShot",
            hist__pb2.SnapShotRequest.SerializeToString,
            hist__pb2.SnapShotResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )

    @staticmethod
    def Flush(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/HistogrammerService/Flush",
            hist__pb2.FlushRequest.SerializeToString,
            hist__pb2.FlushResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )



from __future__ import annotations

import asyncio
from datetime import datetime, timedelta
import logging
from concurrent import futures
import json
import uuid
import threading

import grpc
import hist
import hist.serialization
import numpy as np
import uhi.io.json


from histserv.protos import hist_pb2, hist_pb2_grpc
from histserv.serialize import deserialize, serialize

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("histserv")


class Histogrammer(hist_pb2_grpc.HistogrammerServiceServicer):
    def __init__(self) -> None:
        super().__init__()
        self._lock = threading.Lock()
        self._hists = {}
        self._hists_last_access = {}

    @property
    def hists(self) -> dict:
        return self._hists

    @property
    def hists_last_access(self) -> dict:
        return self._hists_last_access

    async def Init(
        self, request: hist_pb2.InitRequest, context: grpc.ServicerContext
    ) -> hist_pb2.InitResponse:
        # generate unique key
        H_id = uuid.uuid4().hex

        # make sure it doesn't exist (should basically never happen)
        if H_id in self.hists:
            msg = f"Try again; init failed due to existing key (key={H_id})."
            logger.error(msg)
            return hist_pb2.InitResponse(success=False, message=msg)

        # deserialize message to histogram
        H = hist.Hist(
            json.loads(request.hist_json, object_hook=uhi.io.json.object_hook)
        )

        with self._lock:
            # store histogram
            self.hists_last_access[H_id] = datetime.now()
            self.hists[H_id] = H

        logger.info(f"Initialized histogram ({H_id})")

        # return H_id to client to access it again
        return hist_pb2.InitResponse(
            success=True,
            message=H_id,
        )

    async def Fill(
        self, request: hist_pb2.FillRequest, context: grpc.ServicerContext
    ) -> hist_pb2.FillResponse:
        del context  # unused

        H_id = request.hist_id

        try:
            # Deserialize the message from the request
            kwargs = {key: deserialize(msg) for key, msg in request.kwargs.items()}
        except Exception as e:
            msg = f"Error deserializing request: {e!r}"
            logger.error(msg)
            return hist_pb2.FillResponse(success=False, message=msg)
        try:
            with self._lock:
                self.hists_last_access[H_id] = datetime.now()
                self.hists[H_id].fill(**kwargs)
                nbytes_filled = sum(
                    [nd.nbytes for nd in kwargs.values() if isinstance(nd, np.ndarray)]
                )
                logger.info(f"Filled histogram ({H_id}) with {nbytes_filled:,} bytes")
            return hist_pb2.FillResponse(success=True, message=None)
        except Exception as e:
            msg = f"Error filling histogram ({H_id}): {e!r}"
            logger.error(msg)
            return hist_pb2.FillResponse(success=False, message=msg)

    async def SnapShot(
        self, request: hist_pb2.SnapShotRequest, context: grpc.ServicerContext
    ) -> hist_pb2.SnapShotResponse:
        del context  # unused

        H_id = request.hist_id

        try:
            with self._lock:
                self.hists_last_access[H_id] = datetime.now()
                if request.drop_from_server:
                    H = self.hists.pop(H_id)
                else:
                    H = self.hists[H_id]

            # serialize
            H_ser = hist.serialization.to_uhi(H)
            storage = H_ser.pop("storage")
            # recover type
            H_ser["storage"] = {"type": storage.pop("type")}

            # serialize all contents
            data_ser = {k: serialize(v) for k, v in storage.items()}

            logger.info(f"Created snapshot of histogram ({H_id})")

            return hist_pb2.SnapShotResponse(
                success=True,
                message="",
                hist_json=json.dumps(H_ser),  # pure metadata
                data=data_ser,  # heavy contents
            )

        except Exception as e:
            msg = f"Failed creating a snapshot of histogram ({H_id}): {e!r}"
            logger.error(msg)
            return hist_pb2.SnapShotResponse(
                success=False, message=msg, hist_json="", data={}
            )

    async def Flush(
        self, request: hist_pb2.FlushRequest, context: grpc.ServicerContext
    ) -> hist_pb2.FlushResponse:
        import h5py
        import uhi.io.hdf5

        del context  # unused

        H_id = request.hist_id
        destination = request.destination

        if not destination.endswith((".h5", ".hdf5")):
            msg = f"Invalid destination: {destination}, needs to be a hdf5 file, e.g., 'hist.hdf5'"
            logger.error(msg)
            return hist_pb2.FlushResponse(success=False, message=msg)

        try:
            self.hists_last_access[H_id] = datetime.now()
            H = self.hists[H_id]
            with h5py.File(destination, "w") as h5_file:
                uhi.io.hdf5.write(h5_file.create_group(H_id), H)

            with self._lock:
                # flushing means we evict it from the server's memory
                self.hists.pop(H_id, None)
                self.hists_last_access.pop(H_id, None)

            logger.info(f"Flushed histogram ({H_id}) to {destination}")
            return hist_pb2.FlushResponse(
                success=True,
                message=f"Histogram ({H_id}) flushed successfully to {destination}.",
            )
        except Exception as e:
            msg = f"Error flushing histogram ({H_id}): {e!r} to {destination}"
            logger.error(msg)
            return hist_pb2.FlushResponse(success=False, message=msg)


# Callback to remove old hists to make the server memory not explode with long uptimes
async def prune_old_hists(histogrammer: Histogrammer, delta: timedelta) -> None:
    while True:
        now = datetime.now()
        drop_ids = set()
        for H_id, last_access in histogrammer.hists_last_access.items():
            if (now - last_access) >= delta:
                drop_ids.add(H_id)

        with histogrammer._lock:
            for H_id in drop_ids:
                logger.info(
                    f"Dropping histogram ({H_id}) because it hasn't been accessed since {delta}"
                )
                histogrammer.hists_last_access.pop(H_id, None)
                histogrammer.hists.pop(H_id, None)

        # check freq: 5min (todo: make it configurable)
        await asyncio.sleep(60 * 5)


class Server:
    def __init__(
        self,
        port: int = 50051,
        n_threads: int = 1,
    ) -> None:
        self.port = port
        self.n_threads = n_threads

        # create gRPC server
        self.server = grpc.aio.server(
            futures.ThreadPoolExecutor(max_workers=n_threads),
            # compression=grpc.Compression.Gzip,
            compression=grpc.Compression.NoCompression,  # turn off for now, we compress with numcodecs the byte buffer
            options=[
                ("grpc.max_receive_message_length", 1 << 29),
            ],
        )
        # add service
        histogrammer = Histogrammer()
        hist_pb2_grpc.add_HistogrammerServiceServicer_to_server(
            histogrammer, self.server
        )

        # add callbacks:
        # - prune hists that haven't been filled/snapshotted/etc since more than 1 day (todo: make it configurable)
        self._callbacks = [
            asyncio.create_task(
                prune_old_hists(histogrammer=histogrammer, delta=timedelta(days=1))
            ),
        ]

        # add port
        self.server.add_insecure_port(self.address)

    @property
    def address(self) -> str:
        return f"[::]:{self.port}"

    @property
    def callbacks(self) -> list[asyncio.Task]:
        return self._callbacks

    async def start(self) -> None:
        await self.server.start()
        logger.info(
            f"Histogram server started, listening on {self.address} with {self.n_threads} threads"
        )

    async def stop(self, grace: float | None = None) -> None:
        await self.server.stop(grace=grace)

    async def wait_for_termination(self, timeout: float | None = None) -> None:
        await self.server.wait_for_termination(timeout=timeout)



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class InitRequest(_message.Message):
    __slots__ = ("hist_json",)
    HIST_JSON_FIELD_NUMBER: _ClassVar[int]
    hist_json: str
    def __init__(self, hist_json: _Optional[str] = ...) -> None: ...

class InitResponse(_message.Message):
    __slots__ = ("success", "message")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    message: str
    def __init__(self, success: bool = ..., message: _Optional[str] = ...) -> None: ...

class FillRequest(_message.Message):
    __slots__ = ("hist_id", "kwargs")
    class KwargsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: Value
        def __init__(
            self,
            key: _Optional[str] = ...,
            value: _Optional[_Union[Value, _Mapping]] = ...,
        ) -> None: ...

    HIST_ID_FIELD_NUMBER: _ClassVar[int]
    KWARGS_FIELD_NUMBER: _ClassVar[int]
    hist_id: str
    kwargs: _containers.MessageMap[str, Value]
    def __init__(
        self,
        hist_id: _Optional[str] = ...,
        kwargs: _Optional[_Mapping[str, Value]] = ...,
    ) -> None: ...

class FillResponse(_message.Message):
    __slots__ = ("success", "message")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    message: str
    def __init__(self, success: bool = ..., message: _Optional[str] = ...) -> None: ...

class SnapShotRequest(_message.Message):
    __slots__ = ("hist_id", "drop_from_server")
    HIST_ID_FIELD_NUMBER: _ClassVar[int]
    DROP_FROM_SERVER_FIELD_NUMBER: _ClassVar[int]
    hist_id: str
    drop_from_server: bool
    def __init__(
        self, hist_id: _Optional[str] = ..., drop_from_server: bool = ...
    ) -> None: ...

class SnapShotResponse(_message.Message):
    __slots__ = ("success", "message", "hist_json", "data")
    class DataEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: Value
        def __init__(
            self,
            key: _Optional[str] = ...,
            value: _Optional[_Union[Value, _Mapping]] = ...,
        ) -> None: ...

    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    HIST_JSON_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    success: bool
    message: str
    hist_json: str
    data: _containers.MessageMap[str, Value]
    def __init__(
        self,
        success: bool = ...,
        message: _Optional[str] = ...,
        hist_json: _Optional[str] = ...,
        data: _Optional[_Mapping[str, Value]] = ...,
    ) -> None: ...

class FlushRequest(_message.Message):
    __slots__ = ("hist_id", "destination")
    HIST_ID_FIELD_NUMBER: _ClassVar[int]
    DESTINATION_FIELD_NUMBER: _ClassVar[int]
    hist_id: str
    destination: str
    def __init__(
        self, hist_id: _Optional[str] = ..., destination: _Optional[str] = ...
    ) -> None: ...

class FlushResponse(_message.Message):
    __slots__ = ("success", "message")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    message: str
    def __init__(self, success: bool = ..., message: _Optional[str] = ...) -> None: ...

class Dtype(_message.Message):
    __slots__ = ("type",)
    class Type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        TYPE_DT_FLOAT32: _ClassVar[Dtype.Type]
        TYPE_DT_FLOAT64: _ClassVar[Dtype.Type]
        TYPE_DT_INT32: _ClassVar[Dtype.Type]
        TYPE_DT_INT64: _ClassVar[Dtype.Type]

    TYPE_DT_FLOAT32: Dtype.Type
    TYPE_DT_FLOAT64: Dtype.Type
    TYPE_DT_INT32: Dtype.Type
    TYPE_DT_INT64: Dtype.Type
    TYPE_FIELD_NUMBER: _ClassVar[int]
    type: Dtype.Type
    def __init__(self, type: _Optional[_Union[Dtype.Type, str]] = ...) -> None: ...

class Ndarray(_message.Message):
    __slots__ = ("shape", "dtype", "data")
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    shape: _containers.RepeatedScalarFieldContainer[int]
    dtype: Dtype
    data: bytes
    def __init__(
        self,
        shape: _Optional[_Iterable[int]] = ...,
        dtype: _Optional[_Union[Dtype, _Mapping]] = ...,
        data: _Optional[bytes] = ...,
    ) -> None: ...

class Value(_message.Message):
    __slots__ = ("array_value", "string_value", "int_value", "bool_value")
    ARRAY_VALUE_FIELD_NUMBER: _ClassVar[int]
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    INT_VALUE_FIELD_NUMBER: _ClassVar[int]
    BOOL_VALUE_FIELD_NUMBER: _ClassVar[int]
    array_value: Ndarray
    string_value: str
    int_value: int
    bool_value: bool
    def __init__(
        self,
        array_value: _Optional[_Union[Ndarray, _Mapping]] = ...,
        string_value: _Optional[str] = ...,
        int_value: _Optional[int] = ...,
        bool_value: bool = ...,
    ) -> None: ...



from __future__ import annotations

import hist
import numpy as np

from histserv import Client


# some histogram to fill
H = hist.Hist(
    hist.axis.Regular(10, -2, 2, name="x", label="X Axis"),
    hist.axis.Regular(10, -2, 2, name="y", label="Y Axis"),
    hist.axis.StrCategory(
        ["data", "drell-yan"], name="dataset", label="Dataset", growth=True
    ),
    storage=hist.storage.Weight(),
)


if __name__ == "__main__":
    # connect to remote HistServ gRPC server
    with Client(address="[::]:50051") as client:
        # initialize it on the server and receive a 'remote_hist' to interact with it
        remote_hist = client.init(H)

        print("Remote hist initialized:", remote_hist)

        # fill histogram remotely (blocking operation)
        response = remote_hist.fill(
            x=np.random.normal(size=1_000_000).astype(np.float64),
            y=np.random.normal(size=1_000_000).astype(np.float64),
            dataset="data",
            weight=np.ones(1_000_000, dtype=np.float64),
        )
        print(f"Remote hist filled successfully? -> {response.success}")

        # Creating a snapshot means to return the current state of the remote_hist to the client
        # The `drop_from_server` option allows to remove the histogram from the server's memory if set to True
        print(
            "Snapshotting current hist:", remote_hist.snapshot(drop_from_server=False)
        )

        # fill histogram remotely again with different dataset
        response = remote_hist.fill(
            x=np.random.normal(size=1_000_000).astype(np.float64),
            y=np.random.normal(size=1_000_000).astype(np.float64),
            dataset="drell-yan",
            weight=np.ones(1_000_000, dtype=np.float64),
        )
        print(f"Remote hist filled successfully? -> {response.success}")

        # fill histogram remotely again with different dataset (something that triggers axis growth)
        response = remote_hist.fill(
            x=np.random.normal(size=1_000_000).astype(np.float64),
            y=np.random.normal(size=1_000_000).astype(np.float64),
            dataset="ttbar",
            weight=np.ones(1_000_000, dtype=np.float64),
        )
        print(f"Remote hist filled successfully? -> {response.success}")

        # flush histogram remotely to file
        response = remote_hist.flush(destination="hist.h5")
        print(f"Remote hist flushed succesfully? -> {response.message}")



from __future__ import annotations

import asyncio
from argparse import ArgumentParser

from histserv.server import Server, logger

# Coroutines to be invoked when the event loop is shutting down.
_cleanup_coroutines = []


async def main() -> None:
    ap = ArgumentParser()
    ap.add_argument("-p", "--port", default=0, type=int)
    ap.add_argument("-t", "--n-threads", default=1, type=int)

    args = ap.parse_args()

    if not 0 <= args.port < 0xFFFF:
        raise ValueError("port must be between 0 and 65535")

    # start server
    server = Server(
        port=args.port,
        n_threads=args.n_threads,
    )
    await server.start()

    async def server_graceful_shutdown():
        logger.info("Starting graceful shutdown...")

        for task in server.callbacks:
            task.cancel()

        # Shuts down the server with 5 seconds of grace period. During the
        # grace period, the server won't accept new connections and allow
        # existing RPCs to continue within the grace period.
        await server.stop(5)

    _cleanup_coroutines.append(server_graceful_shutdown())
    await server.wait_for_termination()


def run() -> None:
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        loop.run_until_complete(main())
    finally:
        if _cleanup_coroutines:
            loop.run_until_complete(*_cleanup_coroutines)
        loop.close()


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: hist.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 6, 31, 1, "", "hist.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\nhist.proto" \n\x0bInitRequest\x12\x11\n\thist_json\x18\x01 \x01(\t"0\n\x0cInitResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t"\x7f\n\x0b\x46illRequest\x12\x0f\n\x07hist_id\x18\x01 \x01(\t\x12(\n\x06kwargs\x18\x02 \x03(\x0b\x32\x18.FillRequest.KwargsEntry\x1a\x35\n\x0bKwargsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x15\n\x05value\x18\x02 \x01(\x0b\x32\x06.Value:\x02\x38\x01"0\n\x0c\x46illResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t"<\n\x0fSnapShotRequest\x12\x0f\n\x07hist_id\x18\x01 \x01(\t\x12\x18\n\x10\x64rop_from_server\x18\x02 \x01(\x08"\xa7\x01\n\x10SnapShotResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x11\n\thist_json\x18\x03 \x01(\t\x12)\n\x04\x64\x61ta\x18\x04 \x03(\x0b\x32\x1b.SnapShotResponse.DataEntry\x1a\x33\n\tDataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x15\n\x05value\x18\x02 \x01(\x0b\x32\x06.Value:\x02\x38\x01"4\n\x0c\x46lushRequest\x12\x0f\n\x07hist_id\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65stination\x18\x02 \x01(\t"1\n\rFlushResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t"z\n\x05\x44type\x12\x19\n\x04type\x18\x02 \x01(\x0e\x32\x0b.Dtype.Type"V\n\x04Type\x12\x13\n\x0fTYPE_DT_FLOAT32\x10\x00\x12\x13\n\x0fTYPE_DT_FLOAT64\x10\x01\x12\x11\n\rTYPE_DT_INT32\x10\x02\x12\x11\n\rTYPE_DT_INT64\x10\x03"=\n\x07Ndarray\x12\r\n\x05shape\x18\x01 \x03(\x03\x12\x15\n\x05\x64type\x18\x02 \x01(\x0b\x32\x06.Dtype\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c"t\n\x05Value\x12\x1f\n\x0b\x61rray_value\x18\x01 \x01(\x0b\x32\x08.NdarrayH\x00\x12\x16\n\x0cstring_value\x18\x02 \x01(\tH\x00\x12\x13\n\tint_value\x18\x03 \x01(\x03H\x00\x12\x14\n\nbool_value\x18\x04 \x01(\x08H\x00\x42\x07\n\x05value2\xc0\x01\n\x13HistogrammerService\x12%\n\x04Init\x12\x0c.InitRequest\x1a\r.InitResponse"\x00\x12%\n\x04\x46ill\x12\x0c.FillRequest\x1a\r.FillResponse"\x00\x12\x31\n\x08SnapShot\x12\x10.SnapShotRequest\x1a\x11.SnapShotResponse"\x00\x12(\n\x05\x46lush\x12\r.FlushRequest\x1a\x0e.FlushResponse"\x00\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "hist_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_FILLREQUEST_KWARGSENTRY"]._loaded_options = None
    _globals["_FILLREQUEST_KWARGSENTRY"]._serialized_options = b"8\001"
    _globals["_SNAPSHOTRESPONSE_DATAENTRY"]._loaded_options = None
    _globals["_SNAPSHOTRESPONSE_DATAENTRY"]._serialized_options = b"8\001"
    _globals["_INITREQUEST"]._serialized_start = 14
    _globals["_INITREQUEST"]._serialized_end = 46
    _globals["_INITRESPONSE"]._serialized_start = 48
    _globals["_INITRESPONSE"]._serialized_end = 96
    _globals["_FILLREQUEST"]._serialized_start = 98
    _globals["_FILLREQUEST"]._serialized_end = 225
    _globals["_FILLREQUEST_KWARGSENTRY"]._serialized_start = 172
    _globals["_FILLREQUEST_KWARGSENTRY"]._serialized_end = 225
    _globals["_FILLRESPONSE"]._serialized_start = 227
    _globals["_FILLRESPONSE"]._serialized_end = 275
    _globals["_SNAPSHOTREQUEST"]._serialized_start = 277
    _globals["_SNAPSHOTREQUEST"]._serialized_end = 337
    _globals["_SNAPSHOTRESPONSE"]._serialized_start = 340
    _globals["_SNAPSHOTRESPONSE"]._serialized_end = 507
    _globals["_SNAPSHOTRESPONSE_DATAENTRY"]._serialized_start = 456
    _globals["_SNAPSHOTRESPONSE_DATAENTRY"]._serialized_end = 507
    _globals["_FLUSHREQUEST"]._serialized_start = 509
    _globals["_FLUSHREQUEST"]._serialized_end = 561
    _globals["_FLUSHRESPONSE"]._serialized_start = 563
    _globals["_FLUSHRESPONSE"]._serialized_end = 612
    _globals["_DTYPE"]._serialized_start = 614
    _globals["_DTYPE"]._serialized_end = 736
    _globals["_DTYPE_TYPE"]._serialized_start = 650
    _globals["_DTYPE_TYPE"]._serialized_end = 736
    _globals["_NDARRAY"]._serialized_start = 738
    _globals["_NDARRAY"]._serialized_end = 799
    _globals["_VALUE"]._serialized_start = 801
    _globals["_VALUE"]._serialized_end = 917
    _globals["_HISTOGRAMMERSERVICE"]._serialized_start = 920
    _globals["_HISTOGRAMMERSERVICE"]._serialized_end = 1112
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import typing as tp

import numcodecs
import numpy as np

from histserv.protos import hist_pb2

# hardcode for now, make it configurable later
codec = numcodecs.Blosc(cname="zstd", clevel=9, shuffle=numcodecs.Blosc.BITSHUFFLE)


def serialize_nparray(item: np.ndarray) -> hist_pb2.Ndarray:
    assert isinstance(item, np.ndarray)

    shape = list(item.shape)
    dtype = _numpy_dtype_to_proto_dtype(item.dtype)
    compressed_data = codec.encode(item.tobytes())
    return hist_pb2.Ndarray(shape=shape, dtype=dtype, data=compressed_data)


def serialize(item: tp.Any) -> hist_pb2.Value:
    """Serialize something into a hist_pb2.Value message."""
    msg = hist_pb2.Value()
    if isinstance(item, np.ndarray):
        msg.array_value.CopyFrom(serialize_nparray(item))
    elif isinstance(item, str):
        msg.string_value = item
    elif isinstance(item, int):
        msg.int_value = item
    elif isinstance(item, bool):
        msg.bool_value = item
    else:
        raise TypeError(f"Can't serialize: {item}")
    return msg


def _numpy_dtype_to_proto_dtype(np_dtype: np.dtype) -> hist_pb2.Dtype:
    """Mapping from numpy dtypes to hist_pb2.dtype messages."""
    match np_dtype:
        case np.float32:
            return hist_pb2.Dtype(type=hist_pb2.Dtype.TYPE_DT_FLOAT32)
        case np.float64:
            return hist_pb2.Dtype(type=hist_pb2.Dtype.TYPE_DT_FLOAT64)
        case np.int32:
            return hist_pb2.Dtype(type=hist_pb2.Dtype.TYPE_DT_INT32)
        case np.int64:
            return hist_pb2.Dtype(type=hist_pb2.Dtype.TYPE_DT_INT64)
        case _:
            raise ValueError(f"Unsupported numpy dtype: {np_dtype}")


def deserialize(message: hist_pb2.Value):
    """Deserialize a hist_pb2.Value message into a numpy ndarray, str or int."""
    match message.WhichOneof("value"):
        case "array_value":
            ndarray = message.array_value
            shape = tuple(ndarray.shape)
            dtype = _proto_dtype_to_numpy_dtype(ndarray.dtype)
            decompressed_data = codec.decode(ndarray.data)
            array = np.frombuffer(decompressed_data, dtype=dtype).reshape(shape)
            return array
        case "string_value":
            return message.string_value
        case "int_value":
            return message.int_value
        case "bool_value":
            return message.bool_value
        case _:
            raise ValueError("Unsupported Value type")


def _proto_dtype_to_numpy_dtype(proto_dtype: hist_pb2.Dtype):
    """Mapping from hist_pb2.dtype messages to numpy dtypes."""
    match proto_dtype.type:
        case hist_pb2.Dtype.TYPE_DT_FLOAT32:
            return np.float32
        case hist_pb2.Dtype.TYPE_DT_FLOAT64:
            return np.float64
        case hist_pb2.Dtype.TYPE_DT_INT32:
            return np.int32
        case hist_pb2.Dtype.TYPE_DT_INT64:
            return np.int64
        case _:
            raise ValueError(f"Unsupported proto dtype: {proto_dtype.type}")



import hist
from histserv import Client


class Processor:
    def __init__(self, client: Client):
        # setup a histogram to fill
        H = hist.Hist(
            hist.axis.Regular(10, -2, 2, name="x", label="X Axis"),
            hist.axis.Regular(10, -2, 2, name="y", label="Y Axis"),
            hist.axis.StrCategory(
                ["data", "drell-yan"], name="dataset", label="Dataset", growth=True
            ),
            storage=hist.storage.Weight(),
        )

        # initialize it on the server and receive a 'remote_hist' to interact with it
        # do that only once on the host machine, every `init` will initialize a new
        # histogram to fill on the server, but we only want to do that once and fill that one
        self.remote_hist = client.init(H)

    def process(self, events):
        # simple example fill
        import numpy as np

        for _ in range(4):
            ret = self.remote_hist.fill(
                x=np.random.normal(size=1_000_000).astype(np.float64),
                y=np.random.normal(size=1_000_000).astype(np.float64),
                dataset="data",
                weight=np.ones(1_000_000, dtype=np.float64),
            )
            if not ret.success:
                raise ValueError(ret.message)

        print("All remote fills succeeded!")

        # no need to return something, we filled already remotely
        return None


if __name__ == "__main__":
    client = Client(address="[::]:50051")

    myanalysis = Processor(client=client)

    # check pickling works for coffea dask workflows
    import pickle

    myanalysis = pickle.loads(pickle.dumps(myanalysis))

    # run analysis
    myanalysis.process(events=None)

    print("Output hist:", myanalysis.remote_hist.snapshot())
